@@ -9,19 +9,13 @@
       INTEGER, intent(inout) :: nx, ny, this_step
       INTEGER j
       INTEGER i
-      INTEGER istop, jstop
-      !
-      ! Look-up loop bounds
-      istop = fld1%grid%subdomain%internal%xstop
-      jstop = fld1%grid%subdomain%internal%ystop
-      !
       !$acc enter data copyin(fld1,fld1%data,nx,ny,this_step)
       fld1%data_on_device = .true.
       !
       !$acc parallel default(present)
-      !$acc loop collapse(2) independent
-      DO j=2,jstop
-        DO i=2,istop
+      !$acc loop independent collapse(2)
+      DO j=fld1%internal%ystart,fld1%internal%ystop
+        DO i=fld1%internal%xstart,fld1%internal%xstop
           CALL inc_field_0_code(i, j, fld1%data, nx, ny, this_step)
         END DO
       END DO
