@@ -1,7 +1,7 @@
 # -----------------------------------------------------------------------------
 # BSD 3-Clause License
 #
-# Copyright (c) 2018-2020, Science and Technology Facilities Council.
+# Copyright (c) 2018-2021, Science and Technology Facilities Council.
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -31,7 +31,7 @@
 # ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 # POSSIBILITY OF SUCH DAMAGE.
 # -----------------------------------------------------------------------------
-# Authors: R. W. Ford and A. R. Porter, STFC Daresbury Lab
+# Authors: R. W. Ford, A. R. Porter and S. Siso STFC Daresbury Lab
 # Modified: J. Henrichs, Bureau of Meteorology,
 #           I. Kavcic, Met Office
 
@@ -44,6 +44,8 @@ Deals with reading the config file and storing default settings.
 from __future__ import absolute_import, print_function
 from collections import namedtuple
 import os
+import six
+from psyclone.errors import InternalError
 
 
 # Name of the config file we search for
@@ -204,6 +206,9 @@ class Config(object):
         # The root name to use when creating internal PSyIR names.
         self._psyir_root_name = None
 
+        # Number of OpenCL devices per node
+        self._ocl_devices_per_node = 1
+
     # -------------------------------------------------------------------------
     def load(self, config_file=None):
         '''Loads a configuration file.
@@ -225,7 +230,10 @@ class Config(object):
             self._config_file = Config.find_file()
         from configparser import ConfigParser, MissingSectionHeaderError, \
             ParsingError
-        self._config = ConfigParser()
+        # Add a getlist method to the ConfigParser instance using the
+        # converters argument
+        self._config = ConfigParser(
+            converters={'list': lambda x: [i.strip() for i in x.split(',')]})
         try:
             self._config.read(self._config_file)
         # Check for missing section headers and general parsing errors
@@ -255,7 +263,7 @@ class Config(object):
                 'DISTRIBUTED_MEMORY')
         except ValueError as err:
             raise ConfigurationError(
-                "error while parsing DISTRIBUTED_MEMORY: {0}".
+                "Error while parsing DISTRIBUTED_MEMORY: {0}".
                 format(str(err)), config=self)
 
         # API for psyclone
@@ -301,7 +309,7 @@ class Config(object):
                 'REPRODUCIBLE_REDUCTIONS')
         except ValueError as err:
             raise ConfigurationError(
-                "error while parsing REPRODUCIBLE_REDUCTIONS: {0}".
+                "Error while parsing REPRODUCIBLE_REDUCTIONS: {0}".
                 format(str(err)), config=self)
 
         try:
@@ -319,13 +327,19 @@ class Config(object):
         else:
             self._psyir_root_name = self._config['DEFAULT']['PSYIR_ROOT_NAME']
 
-        # Read the valid PSyData class prefixes:
-        try:
-            # Convert it to a list of strings:
-            self._valid_psy_data_prefixes = \
-                self._config["DEFAULT"]["VALID_PSY_DATA_PREFIXES"].split()
-        except KeyError:
+        # Read the valid PSyData class prefixes. If the keyword does
+        # not exist then return an empty list.
+        self._valid_psy_data_prefixes = \
+            self._config["DEFAULT"].getlist("VALID_PSY_DATA_PREFIXES")
+        if self._valid_psy_data_prefixes is None:
             self._valid_psy_data_prefixes = []
+        try:
+            self._ocl_devices_per_node = self._config['DEFAULT'].getint(
+                'OCL_DEVICES_PER_NODE')
+        except ValueError as err:
+            six.raise_from(ConfigurationError(
+                "error while parsing OCL_DEVICES_PER_NODE: "
+                "{0}".format(str(err)), config=self), err)
 
         # Verify that the prefixes will result in valid Fortran names:
         import re
@@ -663,6 +677,12 @@ class Config(object):
         :rtype: list of str'''
         return self._valid_psy_data_prefixes
 
+    @property
+    def ocl_devices_per_node(self):
+        ''':returns: The number of OpenCL devices per node.
+        :rtype: int'''
+        return self._ocl_devices_per_node
+
     def get_default_keys(self):
         '''Returns all keys from the default section.
         :returns list: List of all keys of the default section as strings.
@@ -671,6 +691,7 @@ class Config(object):
 
 
 # =============================================================================
+# TODO (issue #1163): Add support for lists in the configuration file
 class APISpecificConfig(object):
     '''A base class for functions that each API-specific class must provide.
     At the moment this is just the function 'access_mapping' that maps between
@@ -688,12 +709,13 @@ class APISpecificConfig(object):
         self._access_mapping = {"read": "read", "write": "write",
                                 "readwrite": "readwrite", "inc": "inc",
                                 "sum": "sum"}
-        # Get the mapping and convert it into a dictionary. The input is in
-        # the format: key1:value1, key2=value2, ...
-        mapping = section.get("ACCESS_MAPPING")
-        if mapping:
+        # Get the mapping if one exists and convert it into a
+        # dictionary. The input is in the format: key1:value1,
+        # key2=value2, ...
+        mapping_list = section.getlist("ACCESS_MAPPING")
+        if mapping_list is not None:
             self._access_mapping = \
-                APISpecificConfig.create_dict_from_string(mapping)
+                APISpecificConfig.create_dict_from_list(mapping_list)
         # Now convert the string type ("read" etc) to AccessType
         # TODO (issue #710): Add checks for duplicate or missing access
         # key-value pairs
@@ -714,34 +736,32 @@ class APISpecificConfig(object):
                                         self._access_mapping.items()}
 
     @staticmethod
-    def create_dict_from_string(input_str):
-        '''Takes an input string in the format:
-        key1:value1, key2:value2, ...
-        and creates a dictionary with the key,value pairs.
-        Spaces are removed.
-        :param str input_str: The input string.
-        :returns: A dictionary with the key,value pairs from the input string.
-        :rtype: dict.
-        :raises ConfigurationError: if the input string contains an entry \
-                that does not have a ":".
-        '''
-        # Remove spaces and convert unicode to normal strings.
-        input_str = str(input_str.strip())
-        if not input_str:
-            # Split will otherwise return a list with '' as only element,
-            # which then raises an exception
-            return {}
+    def create_dict_from_list(input_list):
+        '''Takes a list of strings each with the format: key:value and creates
+        a dictionary with the key,value pairs. Any leading or trailing
+        white space is removed.
 
-        entries = input_str.split(",")
+        :param input_list: the input list.
+        :type input_list: list of str
+
+        :returns: a dictionary with the key,value pairs from the input \
+            list.
+        :rtype: dict.
+
+        :raises ConfigurationError: if any entry in the input list \
+                does not contain a ":".
+
+        '''
         return_dict = {}
-        for entry in entries:
+        for entry in input_list:
             try:
                 key, value = entry.split(":", 1)
             except ValueError:
                 # Raised when split does not return two elements:
                 raise ConfigurationError("Invalid format for mapping: {0}".
                                          format(entry.strip()))
-            return_dict[key.strip()] = value.strip()
+            # Remove spaces and convert unicode to normal strings in Python2
+            return_dict[str(key.strip())] = str(value.strip())
         return return_dict
 
     def get_access_mapping(self):
@@ -775,14 +795,31 @@ class APISpecificConfig(object):
 # =============================================================================
 class DynConfig(APISpecificConfig):
     '''
-    Dynamo0.3-specific Config sub-class. Holds configuration options specific
-    to the Dynamo 0.3 API.
+    LFRic-specific (Dynamo 0.3) Config sub-class. Holds configuration options
+    specific to the LFRic (Dynamo 0.3) API and Dynamo 0.1 API.
 
-    :param config: The 'parent' Config object.
+    :param config: the 'parent' Config object.
     :type config: :py:class:`psyclone.configuration.Config`
-    :param section: The entry for the dynamo0.3 section of \
+    :param section: the entry for the '[dynamo0.3]' section of \
                     the configuration file, as produced by ConfigParser.
-    :type section:  :py:class:`configparser.SectionProxy`
+    :type section: :py:class:`configparser.SectionProxy`
+
+    :raises ConfigurationError: for a missing mandatory configuration option.
+    :raises ConfigurationError: for an invalid option for the redundant \
+                                computation over annexed dofs.
+    :raises ConfigurationError: for an invalid run_time_checks flag.
+    :raises ConfigurationError: if argument datatypes in the 'default_kind' \
+                                mapping do not match the supported datatypes.
+    :raises ConfigurationError: for an invalid argument kind.
+    :raises ConfigurationError: for an invalid value type of NUM_ANY_SPACE.
+    :raises ConfigurationError: if the supplied number of ANY_SPACE \
+                                function spaces is less than or equal to 0.
+    :raises ConfigurationError: for an invalid value type of \
+                                NUM_ANY_DISCONTINUOUS_SPACE.
+    :raises ConfigurationError: if the supplied number of \
+                                ANY_DISCONTINUOUS_SPACE function \
+                                spaces is less than or equal to 0.
+
     '''
     # pylint: disable=too-few-public-methods
     def __init__(self, config, section):
@@ -794,7 +831,11 @@ class DynConfig(APISpecificConfig):
         # Initialise run_time_checks setting
         self._run_time_checks = None
         # Initialise LFRic datatypes' default kinds (precisions) settings
+        self._supported_fortran_datatypes = []
         self._default_kind = {}
+        # Number of ANY_SPACE and ANY_DISCONTINUOUS_SPACE function spaces
+        self._num_any_space = None
+        self._num_any_discontinuous_space = None
         # Set mandatory keys
         # TODO: to be fully populated here for LFRic (Dynamo0.3) API in #282
         # when Dynamo0.1 API is removed.
@@ -805,60 +846,114 @@ class DynConfig(APISpecificConfig):
         if section.name == "dynamo0.3":
             # Define and check mandatory keys
             self._mandatory_keys = ["access_mapping",
-                                    "compute_annexed_dofs", "default_kind",
-                                    "run_time_checks"]
+                                    "compute_annexed_dofs",
+                                    "supported_fortran_datatypes",
+                                    "default_kind",
+                                    "run_time_checks",
+                                    "num_any_space",
+                                    "num_any_discontinuous_space"]
             mdkeys = set(self._mandatory_keys)
             if not mdkeys.issubset(set(section.keys())):
                 raise ConfigurationError(
-                    "Missing mandatory configuration option in the "
-                    "'[dynamo0.3]' section of the configuration file '{0}'. "
-                    "Valid options are: '{1}'."
-                    .format(config.filename, self._mandatory_keys))
+                    "Missing mandatory configuration option in the '[{0}]' "
+                    "section of the configuration file '{1}'. "
+                    "Valid options are: {2}."
+                    .format(section.name, config.filename,
+                            self._mandatory_keys))
 
             # Parse setting for redundant computation over annexed dofs
             try:
                 self._compute_annexed_dofs = section.getboolean(
                     "compute_annexed_dofs")
             except ValueError as err:
-                raise ConfigurationError(
-                    "error while parsing COMPUTE_ANNEXED_DOFS in the "
-                    "[dynamo0.3] section of the config file: {0}"
-                    .format(str(err)), config=self._config)
+                six.raise_from(ConfigurationError(
+                    "Error while parsing COMPUTE_ANNEXED_DOFS in the "
+                    "'[{0}]' section of the configuration file '{1}': {2}."
+                    .format(section.name, config.filename, str(err)),
+                    config=self._config), err)
 
-            # Setting for run_time_checks flag
+            # Parse setting for run_time_checks flag
             try:
                 self._run_time_checks = section.getboolean(
                     "run_time_checks")
             except ValueError as err:
-                raise ConfigurationError(
-                    "error while parsing RUN_TIME_CHECKS in the "
-                    "[dynamo0.3] section of the config file: {0}"
-                    .format(str(err)), config=self._config)
+                six.raise_from(ConfigurationError(
+                    "Error while parsing RUN_TIME_CHECKS in the '[{0}]' "
+                    "section of the configuration file '{1}': {2}."
+                    .format(section.name, config.filename, str(err)),
+                    config=self._config), err)
 
-            # Parse setting for default kinds (precisions)
-            all_kinds = self.create_dict_from_string(section["default_kind"])
+            # Parse setting for the supported Fortran datatypes. No
+            # need to check whether the keyword is found as it is
+            # mandatory (and therefore already checked).
+            self._supported_fortran_datatypes = section.getlist(
+                "supported_fortran_datatypes")
+
+            # Parse setting for default kinds (precisions). No need to
+            # check whether the keyword is found as it is mandatory
+            # (and therefore already checked).
+            kind_list = section.getlist("default_kind")
+            all_kinds = self.create_dict_from_list(kind_list)
             # Set default kinds (precisions) from config file
-            from psyclone.dynamo0p3 import SUPPORTED_FORTRAN_DATATYPES
             # Check for valid datatypes (filter to remove empty values)
             datatypes = set(filter(None, all_kinds.keys()))
-            if datatypes != set(SUPPORTED_FORTRAN_DATATYPES):
+            if datatypes != set(self._supported_fortran_datatypes):
                 raise ConfigurationError(
-                    "Invalid datatype found in the '[dynamo0.3]' "
-                    "section of the configuration file '{0}'. "
-                    "Valid datatypes are: '{1}'."
-                    .format(config.filename, SUPPORTED_FORTRAN_DATATYPES))
+                    "Fortran datatypes in the 'default_kind' mapping in the "
+                    "'[{0}]' section of the configuration file '{1}' do "
+                    "not match the supported Fortran datatypes {2}."
+                    .format(section.name, config.filename,
+                            self._supported_fortran_datatypes))
             # Check for valid kinds (filter to remove any empty values)
             datakinds = set(filter(None, all_kinds.values()))
-            if len(datakinds) != len(set(SUPPORTED_FORTRAN_DATATYPES)):
+            if len(datakinds) != len(set(self._supported_fortran_datatypes)):
                 raise ConfigurationError(
-                    "Supplied kind parameters '{0}' in the '[dynamo0.3]' "
-                    "section of the configuration file '{1}' do not define "
-                    "the default kind for one or more supported "
-                    "datatypes '{2}'."
-                    .format(sorted(datakinds), config.filename,
-                            SUPPORTED_FORTRAN_DATATYPES))
+                    "Supplied kind parameters {0} in the '[{1}]' section "
+                    "of the configuration file '{2}' do not define the "
+                    "default kind for one or more supported datatypes {3}."
+                    .format(sorted(datakinds), section.name, config.filename,
+                            self._supported_fortran_datatypes))
             self._default_kind = all_kinds
 
+            # Parse setting for the number of ANY_SPACE function spaces
+            # (check for an invalid value and numbers <= 0)
+            try:
+                self._num_any_space = section.getint("NUM_ANY_SPACE")
+            except ValueError as err:
+                six.raise_from(ConfigurationError(
+                    "Error while parsing NUM_ANY_SPACE in the '[{0}]' "
+                    "section of the configuration file '{1}': {2}."
+                    .format(section.name, config.filename, str(err)),
+                    config=self._config), err)
+
+            if self._num_any_space <= 0:
+                raise ConfigurationError(
+                    "The supplied number of ANY_SPACE function spaces "
+                    "in the '[{0}]' section of the configuration "
+                    "file '{1}' must be greater than 0 but found {2}."
+                    .format(section.name, config.filename,
+                            self._num_any_space))
+
+            # Parse setting for the number of ANY_DISCONTINUOUS_SPACE
+            # function spaces (checks for an invalid value and numbers <= 0)
+            try:
+                self._num_any_discontinuous_space = section.getint(
+                    "NUM_ANY_DISCONTINUOUS_SPACE")
+            except ValueError as err:
+                six.raise_from(ConfigurationError(
+                    "Error while parsing NUM_ANY_DISCONTINUOUS_SPACE in the "
+                    "'[{0}]' section of the configuration file '{1}': {2}."
+                    .format(section.name, config.filename, str(err)),
+                    config=self._config), err)
+
+            if self._num_any_discontinuous_space <= 0:
+                raise ConfigurationError(
+                    "The supplied number of ANY_DISCONTINUOUS_SPACE function "
+                    "spaces in the '[{0}]' section of the configuration "
+                    "file '{1}' must be greater than 0 but found {2}."
+                    .format(section.name, config.filename,
+                            self._num_any_discontinuous_space))
+
     @property
     def compute_annexed_dofs(self):
         '''
@@ -882,6 +977,17 @@ class DynConfig(APISpecificConfig):
         '''
         return self._run_time_checks
 
+    @property
+    def supported_fortran_datatypes(self):
+        '''
+        Getter for the supported Fortran argument datatypes in LFRic.
+
+        :returns: supported Fortran datatypes for LFRic arguments.
+        :rtype: list of str
+
+        '''
+        return self._supported_fortran_datatypes
+
     @property
     def default_kind(self):
         '''
@@ -890,9 +996,29 @@ class DynConfig(APISpecificConfig):
 
         :returns: the default kinds for main datatypes in LFRic.
         :rtype: dict of str
+
         '''
         return self._default_kind
 
+    @property
+    def num_any_space(self):
+        '''
+        :returns: the number of ANY_SPACE function spaces in LFRic.
+        :rtype: int
+
+        '''
+        return self._num_any_space
+
+    @property
+    def num_any_discontinuous_space(self):
+        '''
+        :returns: the number of ANY_DISCONTINUOUS_SPACE function \
+                  spaces in LFRic.
+        :rtype: int
+
+        '''
+        return self._num_any_discontinuous_space
+
 
 # =============================================================================
 class GOceanConfig(APISpecificConfig):
@@ -917,6 +1043,8 @@ class GOceanConfig(APISpecificConfig):
         # a property, and 'type' is a string.
         # These values are taken from the psyclone config file.
         self._grid_properties = {}
+        # Initialise debug_mode settings (default to False if it doesn't exist)
+        self._debug_mode = False
         for key in section.keys():
             # Do not handle any keys from the DEFAULT section
             # since they are handled by Config(), not this class.
@@ -935,23 +1063,35 @@ class GOceanConfig(APISpecificConfig):
             elif key == "access_mapping":
                 # Handled in the base class APISpecificConfig
                 pass
+            elif key == "debug_mode":
+                # Boolean that specifies if debug mode is enabled
+                try:
+                    self._debug_mode = section.getboolean("debug_mode")
+                except ValueError as err:
+                    six.raise_from(ConfigurationError(
+                        "error while parsing DEBUG_MODE in the "
+                        "[gocean1p0] section of the config file: {0}"
+                        .format(str(err))), err)
             elif key == "grid-properties":
                 # Grid properties have the format:
-                # go_grid_area_u: {0}%%grid%%area_u: array,
+                # go_grid_area_u: {0}%%grid%%area_u: array: real,
                 # First the name, then the Fortran code to access the property,
-                # followed by the type ("array" or "scalar")
-                all_props = self.create_dict_from_string(section[key])
+                # followed by the type ("array" or "scalar") and then the
+                # intrinsic type ("integer" or "real")
+                all_props = self.create_dict_from_list(section.getlist(key))
                 for grid_property in all_props:
                     try:
-                        fortran, variable_type = all_props[grid_property]\
-                                            .split(":")
+                        fortran, variable_type, intrinsic_type = \
+                            all_props[grid_property].split(":")
                     except ValueError:
                         # Raised when the string does not contain exactly
-                        # two values separated by ":"
+                        # three values separated by ":"
                         error = "Invalid property \"{0}\" found with value " \
                                 "\"{1}\" in \"{2}\". It must have exactly " \
-                                "two ':'-delimited separated values: the " \
-                                "property and the type." \
+                                "three ':'-delimited separated values: the " \
+                                "property, whether it is a scalar or an " \
+                                "array, and the intrinsic type (real or " \
+                                "integer)." \
                                 .format(grid_property,
                                         all_props[grid_property],
                                         config.filename)
@@ -960,7 +1100,8 @@ class GOceanConfig(APISpecificConfig):
                     # file might contain
                     self._grid_properties[grid_property] = \
                         GOceanConfig.make_property(fortran.strip(),
-                                                   variable_type.strip())
+                                                   variable_type.strip(),
+                                                   intrinsic_type.strip())
                 # Check that the required values for xstop and ystop
                 # are defined:
                 for required in ["go_grid_xstop", "go_grid_ystop",
@@ -986,7 +1127,7 @@ class GOceanConfig(APISpecificConfig):
 
     # ---------------------------------------------------------------------
     @staticmethod
-    def make_property(dereference_format, type_name):
+    def make_property(dereference_format, type_name, intrinsic_type):
         '''Creates a property (based on namedtuple) for a given Fortran
         code to access a grid property, and the type.
 
@@ -995,29 +1136,44 @@ class GOceanConfig(APISpecificConfig):
             string. E.g. "{0}%whole%xstop").
         :param str type_name: The type of the grid property, must be \
             'scalar' or 'array'.
+        :param str intrinsic_type: The intrinsic type of the grid property, \
+            must be 'integer' or 'real'.
 
         :returns: a namedtuple for a grid property given the Fortran
             statement to access it and the type.
 
         :raises InternalError: if type_name is not 'scalar' or 'array'
+        :raises InternalError: if intrinsic_type is not 'integer' or 'real'
         '''
         if type_name not in ['array', 'scalar']:
-            from psyclone.errors import InternalError
             raise InternalError("Type must be 'array' or 'scalar' but is "
                                 "'{0}'.".format(type_name))
 
-        Property = namedtuple("Property", "fortran type")
-        return Property(dereference_format, type_name)
+        if intrinsic_type not in ['integer', 'real']:
+            raise InternalError("Intrinsic type must be 'integer' or 'real' "
+                                "but is '{0}'.".format(intrinsic_type))
+
+        Property = namedtuple("Property", "fortran type intrinsic_type")
+        return Property(dereference_format, type_name, intrinsic_type)
 
     # ---------------------------------------------------------------------
     @property
     def grid_properties(self):
         ''':returns: the dict containing the grid properties.
-        :rtype: a dict with values of namedtuple("Property","fortran type") \
-            instances.
+        :rtype: a dict with values of \
+            namedtuple("Property","fortran type intrinsic_type") instances.
         '''
         return self._grid_properties
 
+    @property
+    def debug_mode(self):
+        '''
+        :returns: whether we are generating additional debug code.
+        :rtype: bool
+
+        '''
+        return self._debug_mode
+
 
 # =============================================================================
 class NemoConfig(APISpecificConfig):
@@ -1058,7 +1214,7 @@ class NemoConfig(APISpecificConfig):
             # Handle the definition of variables
             if key[:8] == "mapping-":
                 loop_type = key[8:]
-                data = self.create_dict_from_string(section[key])
+                data = self.create_dict_from_list(section.getlist(key))
                 # Make sure the required keys exist:
                 for subkey in ["var", "start", "stop"]:
                     if subkey not in data:
@@ -1083,8 +1239,7 @@ class NemoConfig(APISpecificConfig):
                 self._loop_type_data[loop_type] = data
 
             elif key == "index-order":
-                self._index_order = [loop.strip() for
-                                     loop in section[key].split(",")]
+                self._index_order = section.getlist(key)
 
             else:
                 raise ConfigurationError("Invalid key \"{0}\" found in "
