@@ -176,9 +176,8 @@ class GOceanExtractNode(ExtractNode):
         sym = Symbol("PSyDataType")
         sym_table.add(sym)
 
-        psy_data = sym_table.new_symbol_name(self.add_psydata_class_prefix
-                                             ("psy_data"))
-        sym_table.add(Symbol(psy_data))
+        root_name = self.add_psydata_class_prefix("psy_data")
+        psy_data = sym_table.new_symbol(root_name).name
         var_decl = TypeDeclGen(prog, datatype=self.add_psydata_class_prefix
                                ("PSyDataType"),
                                entity_decls=[psy_data])
@@ -222,9 +221,8 @@ class GOceanExtractNode(ExtractNode):
                 # No derived type, so we can just use the
                 # variable name directly in the driver
                 local_name = var_name
-            unique_local_name = sym_table.new_symbol_name(local_name)
+            unique_local_name = sym_table.new_symbol(local_name).name
             rename_variable[local_name] = unique_local_name
-            sym_table.add(Symbol(unique_local_name))
             local_name = unique_local_name
 
             # TODO: #644 - we need to identify arrays!!
@@ -325,15 +323,16 @@ class GOceanExtractNode(ExtractNode):
                 # local variable name
                 local_name = prop.fortran[last_percent+1:]
                 unique_name = rename_variable.get(local_name, local_name)
-                all_props[name] = GOceanConfig.make_property(unique_name,
-                                                             prop.type)
+                all_props[name] = GOceanConfig.make_property(
+                    unique_name, prop.type, prop.intrinsic_type)
 
         # 2) The property 'grid_data' is a reference to the data on the
         #    grid (i.e. the actual field) , and it is defined as "{0}%data".
         #    This just becomes {0} ('a_fld%data' in the original program
         #    becomes just 'a_fld', and 'a_fld' is declared to be a plain
         #    Fortran 2d-array)
-        all_props["go_grid_data"] = GOceanConfig.make_property("{0}", "array")
+        all_props["go_grid_data"] = GOceanConfig.make_property(
+            "{0}", "array", "real")
 
         # Each kernel caches the argument code, so we also
         # need to clear this cached data to make sure the new
