@@ -40,7 +40,7 @@
 
 from __future__ import absolute_import
 from psyclone.configuration import Config
-from psyclone.psyir.nodes import Loop, Literal, Schedule
+from psyclone.psyir.nodes import Loop, Literal, Schedule, Reference
 from psyclone.psyGen import PSy, Invokes, Invoke, InvokeSchedule, \
     CodedKern, Arguments, KernelArgument
 from psyclone.psyir.symbols import DataSymbol, INTEGER_TYPE
@@ -137,8 +137,6 @@ class GOPSy(PSy):
                               only=["scalar_field_type"]))
         # add in the subroutines for each invocation
         self.invokes.gen_code(psy_module)
-        # inline kernels where requested
-        self.inline(psy_module)
         return psy_module.root
 
 
@@ -226,8 +224,8 @@ class GOInvokeSchedule(InvokeSchedule):
     supply our API-specific factories to the base InvokeSchedule class
     constructor. '''
 
-    def __init__(self, alg_calls, reserved_names=None):
-        InvokeSchedule.__init__(self, GOKernCallFactory,
+    def __init__(self, name, alg_calls, reserved_names=None):
+        InvokeSchedule.__init__(self, name, GOKernCallFactory,
                                 GOBuiltInCallFactory, alg_calls,
                                 reserved_names)
 
@@ -245,9 +243,19 @@ class GOLoop(Loop):
         self.loop_type = loop_type
 
         if self._loop_type == "inner":
-            self._variable_name = "i"
-        elif self._loop_type == "outer":
-            self._variable_name = "j"
+            tag = "inner_loop_idx"
+            suggested_name = "i"
+        elif self.loop_type == "outer":
+            tag = "outer_loop_idx"
+            suggested_name = "j"
+
+        symtab = self.scope.symbol_table
+        try:
+            self.variable = symtab.lookup_with_tag(tag)
+        except KeyError:
+            self.variable = symtab.new_symbol(
+                suggested_name, tag, symbol_type=DataSymbol,
+                datatype=INTEGER_TYPE)
 
         # Pre-initialise the Loop children  # TODO: See issue #440
         self.addchild(Literal("NOT_INITIALISED", INTEGER_TYPE,
@@ -261,9 +269,9 @@ class GOLoop(Loop):
 
         if self.field_space == "every":
             from psyclone.f2pygen import DeclGen
-            from psyclone.psyir.nodes import BinaryOperation, Reference
+            from psyclone.psyir.nodes import BinaryOperation
             dim_var = DeclGen(parent, datatype="INTEGER",
-                              entity_decls=[self._variable_name])
+                              entity_decls=[self.variable.name])
             parent.add(dim_var)
 
             # Update start loop bound
