@@ -124,12 +124,12 @@ class CWriter(PSyIRVisitor):
         result = "{0}{1} = {2};\n".format(self._nindent, lhs, rhs)
         return result
 
-    def array_node(self, node):
-        '''This method is called when an Array instance is found in the PSyIR
-        tree.
+    def arrayreference_node(self, node):
+        '''This method is called when an ArrayReference instance is found
+        in the PSyIR tree.
 
         :param node: An Array PSyIR node.
-        :type node: :py:class:`psyclone.psyir.nodes.Array`
+        :type node: :py:class:`psyclone.psyir.nodes.ArrayReference`
 
         :returns: The C code as a string.
         :rtype: str
@@ -424,7 +424,7 @@ class CWriter(PSyIRVisitor):
         start = self._visit(node.start_expr)
         stop = self._visit(node.stop_expr)
         step = self._visit(node.step_expr)
-        variable_name = node.variable_name
+        variable_name = node.variable.name
 
         self._depth += 1
         body = ""
@@ -456,5 +456,5 @@ class CWriter(PSyIRVisitor):
         for child in node.dir_body:
             result_list.append(self._visit(child))
         self._depth -= 1
-        result_list.append("}}\n")
+        result_list.append("}\n")
         return "".join(result_list)
