@@ -1,7 +1,7 @@
 # -----------------------------------------------------------------------------
 # BSD 3-Clause License
 #
-# Copyright (c) 2017-2020, Science and Technology Facilities Council.
+# Copyright (c) 2017-2021, Science and Technology Facilities Council.
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -36,39 +36,15 @@
 #         J. Henrichs, Bureau of Meteorology
 # -----------------------------------------------------------------------------
 
-''' This module contains the abstract Node implementation.'''
+'''
+This module contains the abstract Node implementation.
 
-import abc
+'''
+import copy
+import six
 from psyclone.psyir.symbols import SymbolError
 from psyclone.errors import GenerationError, InternalError
 
-# Colour map to use when writing Invoke schedule to terminal. (Requires
-# that the termcolor package be installed. If it isn't then output is not
-# coloured.) See https://pypi.python.org/pypi/termcolor for details.
-SCHEDULE_COLOUR_MAP = {"Schedule": "white",
-                       "Loop": "red",
-                       "GlobalSum": "cyan",
-                       "Directive": "green",
-                       "HaloExchange": "blue",
-                       "HaloExchangeStart": "yellow",
-                       "HaloExchangeEnd": "yellow",
-                       "BuiltIn": "magenta",
-                       "CodedKern": "magenta",
-                       "InlinedKern": "magenta",
-                       "PSyData": "green",
-                       "Profile": "green",
-                       "Extract": "green",
-                       "If": "red",
-                       "Assignment": "blue",
-                       "Range": "white",
-                       "Reference": "yellow",
-                       "Operation": "blue",
-                       "Literal": "yellow",
-                       "Return": "yellow",
-                       "CodeBlock": "red",
-                       "Container": "green"}
-
-
 # Default indentation string
 INDENTATION_STRING = "    "
 
@@ -85,16 +61,34 @@ except ImportError:
         Returns the supplied text argument unchanged. This is a swap-in
         replacement for when termcolor.colored is not available.
 
-        :param text: Text to return
-        :type text: string
-        :param _: Fake argument, only required to match interface
-                  provided by termcolor.colored
-        :returns: The supplied text, unchanged
-        :rtype: string
+        :param str text: text to return.
+        :param _: fake argument, only required to match interface \
+                  provided by termcolor.colored.
+
+        :returns: the supplied text, unchanged.
+        :rtype: str
         '''
         return text
 
 
+def _graphviz_digraph_class():
+    '''
+    Wrapper that returns the graphviz Digraph type if graphviz is installed
+    and None otherwise.
+
+    :returns: the graphviz Digraph type or None.
+    :rtype: :py:class:`graphviz.Digraph` or NoneType.
+
+    '''
+    try:
+        import graphviz as gv
+        return gv.Digraph
+    except ImportError:
+        # TODO #11 add a warning to a log file here
+        # silently return if graphviz bindings are not installed
+        return None
+
+
 class ChildrenList(list):
     '''
     Customized list to keep track of the children nodes. It is initialised with
@@ -143,6 +137,65 @@ class ChildrenList(list):
 
             raise GenerationError(errmsg)
 
+    def _check_is_orphan(self, item):
+        '''
+        Checks that the provided item is an orphan (has no parent or the parent
+        was predefined in the constructor but the other end of connection has
+        not finalised until now).
+
+        :param item: object that needs to be validated.
+        :type item: :py:class:`psyclone.psyir.nodes.Node`
+
+        :raises GenerationError: if the given item is not an orphan.
+        :raises GenerationError: if the item had been provided with a parent \
+            argument on its constructor and this operation is trying to \
+            make its parent a different node.
+
+        '''
+        if item.parent and not item.has_constructor_parent:
+            raise GenerationError(
+                "Item '{0}' can't be added as child of '{1}' because it is not"
+                " an orphan. It already has a '{2}' as a parent.".format(
+                    item.coloured_name(False),
+                    self._node_reference.coloured_name(False),
+                    item.parent.coloured_name(False)))
+        if item.parent and item.has_constructor_parent:
+            if item.parent is not self._node_reference:
+                raise GenerationError(
+                    "'{0}' cannot be set as parent of '{1}' because its "
+                    "constructor predefined the parent reference to a "
+                    "different '{2}' node."
+                    .format(self._node_reference.coloured_name(False),
+                            item.coloured_name(False),
+                            item.parent.coloured_name(False)))
+
+    def _set_parent_link(self, node):
+        '''
+        Set parent connection of the given node to this ChildrenList's node.
+
+        :param node: node for which the parent connection need to be updated.
+        :type node: :py:class:`psyclone.psyir.nodes.Node`
+
+        '''
+        # pylint: disable=protected-access
+        node._parent = self._node_reference
+        node._has_constructor_parent = False
+
+    @staticmethod
+    def _del_parent_link(node):
+        '''
+        Delete parent connection of the given node.
+
+        :param node: node for which the parent connection need to be updated.
+        :type node: :py:class:`psyclone.psyir.nodes.Node`
+
+        '''
+        # This is done from here with protected access because it's the parent
+        # which is in charge of maintaining its children connections.
+        # pylint: disable=protected-access
+        node._parent = None
+        node._has_constructor_parent = False
+
     def append(self, item):
         ''' Extends list append method with children node validation.
 
@@ -151,7 +204,9 @@ class ChildrenList(list):
 
         '''
         self._validate_item(len(self), item)
+        self._check_is_orphan(item)
         super(ChildrenList, self).append(item)
+        self._set_parent_link(item)
 
     def __setitem__(self, index, item):
         ''' Extends list __setitem__ method with children node validation.
@@ -162,7 +217,10 @@ class ChildrenList(list):
 
         '''
         self._validate_item(index, item)
+        self._check_is_orphan(item)
+        self._del_parent_link(self[index])
         super(ChildrenList, self).__setitem__(index, item)
+        self._set_parent_link(item)
 
     def insert(self, index, item):
         ''' Extends list insert method with children node validation.
@@ -174,10 +232,12 @@ class ChildrenList(list):
         '''
         positiveindex = index if index >= 0 else len(self) - index
         self._validate_item(positiveindex, item)
+        self._check_is_orphan(item)
         # Check that all displaced items will still in valid positions
         for position in range(positiveindex, len(self)):
             self._validate_item(position + 1, self[position])
         super(ChildrenList, self).insert(index, item)
+        self._set_parent_link(item)
 
     def extend(self, items):
         ''' Extends list extend method with children node validation.
@@ -188,7 +248,10 @@ class ChildrenList(list):
         '''
         for index, item in enumerate(items):
             self._validate_item(len(self) + index, item)
+            self._check_is_orphan(item)
         super(ChildrenList, self).extend(items)
+        for item in items:
+            self._set_parent_link(item)
 
     # Methods below don't insert elements but have the potential to displace
     # or change the order of the items in-place.
@@ -201,6 +264,7 @@ class ChildrenList(list):
         positiveindex = index if index >= 0 else len(self) - index
         for position in range(positiveindex + 1, len(self)):
             self._validate_item(position - 1, self[position])
+        self._del_parent_link(self[index])
         super(ChildrenList, self).__delitem__(index)
 
     def remove(self, item):
@@ -212,6 +276,7 @@ class ChildrenList(list):
         '''
         for position in range(self.index(item) + 1, len(self)):
             self._validate_item(position - 1, self[position])
+        self._del_parent_link(item)
         super(ChildrenList, self).remove(item)
 
     def pop(self, index=-1):
@@ -225,8 +290,10 @@ class ChildrenList(list):
 
         '''
         positiveindex = index if index >= 0 else len(self) - index
-        for position in range(positiveindex, len(self)):
+        # Check if displaced items after 'positiveindex' will still be valid
+        for position in range(positiveindex + 1, len(self)):
             self._validate_item(position - 1, self[position])
+        self._del_parent_link(self[index])
         return super(ChildrenList, self).pop(index)
 
     def reverse(self):
@@ -270,13 +337,18 @@ class Node(object):
     # properties for each of them and chain the ABC @abstractmethod annotation.
     _children_valid_format = None
     _text_name = None
-    _colour_key = None
+    _colour = None
 
     def __init__(self, ast=None, children=None, parent=None, annotations=None):
         self._children = ChildrenList(self, self._validate_child,
                                       self._children_valid_format)
         if children:
             self._children.extend(children)
+        # Keep a record of whether a parent node was supplied when constructing
+        # this object. In this case it still won't appear in the parent's
+        # children list. When both ends of the reference are connected this
+        # will become False.
+        self._has_constructor_parent = parent is not None
         self._parent = parent
         # Reference into fparser2 AST (if any)
         self._ast = ast
@@ -336,11 +408,19 @@ class Node(object):
                 "given a string value in the concrete class '{0}'."
                 "".format(type(self).__name__))
         if colour:
+            if self._colour is None:
+                raise NotImplementedError(
+                    "The _colour attribute is abstract so needs to be given "
+                    "a string value in the concrete class '{0}'."
+                    "".format(type(self).__name__))
             try:
-                return colored(self._text_name,
-                               SCHEDULE_COLOUR_MAP[self._colour_key])
-            except KeyError:
-                pass
+                return colored(self._text_name, self._colour)
+            except KeyError as info:
+                message = (
+                    "The _colour attribute in class '{0}' has been set to a "
+                    "colour ('{1}') that is not supported by the termcolor "
+                    "package.".format(type(self).__name__, self._colour))
+                six.raise_from(InternalError(message), info)
         return self._text_name
 
     def node_str(self, colour=True):
@@ -354,7 +434,6 @@ class Node(object):
         '''
         return self.coloured_name(colour) + "[]"
 
-    @abc.abstractmethod
     def __str__(self):
         return self.node_str(False)
 
@@ -430,21 +509,34 @@ class Node(object):
         return self._annotations
 
     def dag(self, file_name='dag', file_format='svg'):
-        '''Create a dag of this node and its children.'''
+        '''
+        Create a dag of this node and its children, write it to file and
+        return the graph object.
+
+        :param str file_name: name of the file to create.
+        :param str file_format: format of the file to create. (Must be one \
+                                recognised by Graphviz.)
+
+        :returns: the graph object or None (if the graphviz bindings are not \
+                  installed).
+        :rtype: :py:class:`graphviz.Digraph` or NoneType
+
+        :raises GenerationError: if the specified file format is not \
+                                 recognised by Graphviz.
+
+        '''
+        digraph = _graphviz_digraph_class()
+        if digraph is None:
+            return None
         try:
-            import graphviz as gv
-        except ImportError:
-            # TODO: #11 add a warning to a log file here
-            # silently return if graphviz bindings are not installed
-            return
-        try:
-            graph = gv.Digraph(format=file_format)
+            graph = digraph(format=file_format)
         except ValueError:
             raise GenerationError(
                 "unsupported graphviz file format '{0}' provided".
                 format(file_format))
         self.dag_gen(graph)
         graph.render(filename=file_name)
+        return graph
 
     def dag_gen(self, graph):
         '''Output my node's graph (dag) information and call any
@@ -756,28 +848,35 @@ class Node(object):
         ''' Set a new children list.
 
         :param my_children: new list of children.
-        :type my_children: list or NoneType
+        :type my_children: list
 
-        :raises TypeError: if the given children parameter is not a list \
-            nor NoneType.
+        :raises TypeError: if the given children parameter is not a list.
         '''
-        if my_children is None:
-            self._children = None
-        elif isinstance(my_children, list):
+        if isinstance(my_children, list):
+            self.pop_all_children()  # First remove existing children if any
             self._children = ChildrenList(self, self._validate_child,
                                           self._children_valid_format)
             self._children.extend(my_children)
         else:
             raise TypeError("The 'my_children' parameter of the node.children"
-                            " setter must be a list or None.")
+                            " setter must be a list.")
 
     @property
     def parent(self):
+        '''
+        :returns: the parent node.
+        :rtype: :py:class:`psyclone.psyir.nodes.Node` or NoneType
+        '''
         return self._parent
 
-    @parent.setter
-    def parent(self, my_parent):
-        self._parent = my_parent
+    @property
+    def has_constructor_parent(self):
+        '''
+        :returns: whether the constructor has predefined a parent connection
+            but the parent's children list doesn't include this node yet.
+        :rtype: bool
+        '''
+        return self._has_constructor_parent
 
     @property
     def position(self):
@@ -840,7 +939,17 @@ class Node(object):
 
     @property
     def root(self):
-        node = self
+        '''
+        :returns: the root node of the PSyIR tree.
+        :rtype: :py:class:`psyclone.psyir.nodes.Node`
+
+        '''
+        # Starting with 'self.parent' instead of 'node = self' avoids many
+        # false positive pylint issues that assume self.root type would be
+        # the same as self type.
+        if self.parent is None:
+            return self
+        node = self.parent
         while node.parent is not None:
             node = node.parent
         return node
@@ -870,7 +979,7 @@ class Node(object):
         :param my_type: the class(es) for which the instances are collected.
         :type my_type: either a single :py:class:`psyclone.Node` class \
             or a tuple of such classes
-        :param stop_type: class(es) at which recursion is halted (optional)."
+        :param stop_type: class(es) at which recursion is halted (optional).
 
         :type stop_type: None or a single :py:class:`psyclone.Node` \
             class or a tuple of such classes
@@ -1024,13 +1133,16 @@ class Node(object):
             return True
         return False
 
-    def gen_code(self, parent):
-        '''Abstract base class for code generation function.
-
-        :param parent: the parent of this Node in the PSyIR.
-        :type parent: :py:class:`psyclone.psyir.nodes.Node`
+    def lower_to_language_level(self):
         '''
-        raise NotImplementedError("Please implement me")
+        In-place replacement of DSL or high-level concepts into generic
+        PSyIR constructs. The generic implementation only recurses down
+        to its children, but this method must be re-implemented by Nodes
+        that represent high-level concepts.
+
+        '''
+        for child in self.children:
+            child.lower_to_language_level()
 
     def update(self):
         ''' By default we assume there is no need to update the existing
@@ -1071,136 +1183,139 @@ class Node(object):
             # If we have children then set the Schedule's AST pointer to
             # point to the AST associated with them.
             sched.ast = children[0].ast
-            for child in children:
-                child.parent = sched
         else:
             sched.ast = ast
         return sched
 
-    def find_symbol_table(self):
-        '''
-        :returns: the symbol table attached to the nearest ancestor \
-            node (including self).
-        :rtype: :py:class:`psyclone.psyir.symbols.SymbolTable`
+    @property
+    def scope(self):
+        ''' Some nodes (e.g. Schedule and Container) allow symbols to be
+        scoped via an attached symbol table. This property returns the closest
+        ScopingNode node including self.
 
-        :raises InternalError: if no symbol table is found.
+        :returns: the closest ancestor ScopingNode node.
+        :rtype: :py:class:`psyclone.psyir.node.ScopingNode`
+
+        :raises SymbolError: if there is no ScopingNode ancestor.
 
         '''
-        current = self
-        while current and not hasattr(current, "symbol_table"):
-            current = current.parent
-        if current:
-            return current.symbol_table
-        raise InternalError("Symbol table not found in any ancestor nodes.")
-
-    def find_or_create_symbol(self, name, scope_limit=None):
-        '''Returns the symbol with the name 'name' from a symbol table
-        associated with this node or one of its ancestors.  If the symbol
-        is not found and there are no ContainerSymbols with wildcard imports
-        then an exception is raised. However, if there are one or more
-        ContainerSymbols with wildcard imports (which could therefore be
-        bringing the symbol into scope) then a new DataSymbol of unknown type
-        and interface is created and inserted in the most local SymbolTable
-        that has such an import.
-        The scope_limit variable further limits the symbol table search so
-        that the search through ancestor nodes stops when the scope_limit node
-        is reached i.e. ancestors of the scope_limit node are not searched.
-
-        :param str name: the name of the symbol.
-        :param scope_limit: optional Node which limits the symbol \
-            search space to the symbol tables of the nodes within the \
-            given scope. If it is None (the default), the whole \
-            scope (all symbol tables in ancestor nodes) is searched \
-            otherwise ancestors of the scope_limit node are not \
-            searched.
-        :type scope_limit: :py:class:`psyclone.psyir.nodes.Node` or \
-            `NoneType`
-
-        :returns: the matching symbol.
-        :rtype: :py:class:`psyclone.psyir.symbols.Symbol`
-
-        :raises SymbolError: if no matching symbol is found and there are \
-            no ContainerSymbols from which it might be brought into scope.
-
-        '''
-        from psyclone.psyir.symbols import DataSymbol, UnresolvedInterface, \
-            DeferredType
-        if scope_limit:
-            # Validate the supplied scope_limit
-            if not isinstance(scope_limit, Node):
-                raise TypeError(
-                    "The scope_limit argument '{0}' provided to the "
-                    "find_or_create_symbol method, is not of type `Node`."
-                    "".format(str(scope_limit)))
-
-            # Check that the scope_limit Node is an ancestor of this
-            # Reference Node and raise an exception if not.
-            mynode = self.parent
-            while mynode is not None:
-                if mynode is scope_limit:
-                    # The scope_limit node is an ancestor of the
-                    # supplied node.
-                    break
-                mynode = mynode.parent
-            else:
-                # The scope_limit node is not an ancestor of the
-                # supplied node so raise an exception.
-                raise ValueError(
-                    "The scope_limit node '{0}' provided to the "
-                    "find_or_create_symbol method, is not an ancestor of this "
-                    "node '{1}'.".format(str(scope_limit), str(self)))
-
-        # Keep a list of (symbol table, container) tuples for containers that
-        # have wildcard imports into the current scope and therefore may
-        # contain the symbol we're searching for.
-        possible_containers = []
-        # Keep a reference to the most local SymbolTable with a wildcard
-        # import in case we need to create a Symbol.
-        first_symbol_table = None
-        test_node = self
-
-        # Iterate over ancestor Nodes of this Node.
-        while test_node:
-            # For simplicity, test every Node for the existence of a
-            # SymbolTable (rather than checking for the particular
-            # Node types which we know to have SymbolTables).
-            if hasattr(test_node, 'symbol_table'):
-                # This Node does have a SymbolTable.
-                symbol_table = test_node.symbol_table
-
-                try:
-                    # If the reference matches a Symbol in this
-                    # SymbolTable then return the Symbol.
-                    return symbol_table.lookup(name)
-                except KeyError:
-                    # The Reference Node does not match any Symbols in
-                    # this SymbolTable. Does this SymbolTable have any
-                    # wildcard imports?
-                    for csym in symbol_table.containersymbols:
-                        if csym.wildcard_import:
-                            possible_containers.append((symbol_table, csym))
-                            if not first_symbol_table:
-                                first_symbol_table = symbol_table
-
-            if test_node is scope_limit:
-                # The ancestor scope/top-level Node has been reached and
-                # nothing has matched.
-                break
-
-            # Move on to the next ancestor.
-            test_node = test_node.parent
-
-        if possible_containers:
-            # No symbol found but there are one or more Containers from which
-            # it may be being brought into scope. Therefore create a DataSymbol
-            # of unknown type and deferred interface and add it to the most
-            # local SymbolTable.
-            symbol = DataSymbol(name, DeferredType(),
-                                interface=UnresolvedInterface())
-            first_symbol_table.add(symbol)
-            return symbol
-
-        # All requested Nodes have been checked but there has been no
-        # match and there are no wildcard imports so raise an exception.
+        # These imports have to be local to this method to avoid circular
+        # dependencies.
+        # pylint: disable=import-outside-toplevel
+        from psyclone.psyir.nodes.scoping_node import ScopingNode
+        node = self.ancestor(ScopingNode, include_self=True)
+        if node:
+            return node
         raise SymbolError(
-            "No Symbol found for name '{0}'.".format(name))
+            "Unable to find the scope of node '{0}' as none of its ancestors "
+            "are Container or Schedule nodes.".format(self))
+
+    def replace_with(self, node):
+        '''Removes self, and its descendants, from the PSyIR tree to which it
+        is connected, and replaces it with the supplied node (and its
+        descendants).
+
+        :param node: the node that will replace self in the PSyIR \
+            tree.
+        :type node: :py:class:`psyclone.psyir.nodes.node`
+
+        :raises TypeError: if the argument 'node' is not a Node.
+        :raises GenerationError: if this node does not have a parent.
+        :raises GenerationError: if the argument 'node' has a parent.
+
+        '''
+        if not isinstance(node, Node):
+            raise TypeError(
+                "The argument node in method replace_with in the Node class "
+                "should be a Node but found '{0}'."
+                "".format(type(node).__name__))
+        if not self.parent:
+            raise GenerationError(
+                "This node should have a parent if its replace_with method "
+                "is called.")
+        if node.parent is not None:
+            raise GenerationError(
+                "The parent of argument node in method replace_with in the "
+                "Node class should be None but found '{0}'."
+                "".format(type(node.parent).__name__))
+
+        self.parent.children[self.position] = node
+
+    def pop_all_children(self):
+        ''' Remove all children of this node and return them as a list.
+
+        :returns: all the children of this node as orphan nodes.
+        :rtype: list of :py:class:`psyclone.psyir.node.Node`
+
+        '''
+        free_children = []
+        while self.children:
+            free_children.insert(0, self.children.pop())
+        return free_children
+
+    def detach(self):
+        ''' Detach this node from the tree it belongs to. This is necessary
+        as a precursor to inserting it as the child of another node.
+
+        :returns: this node detached from its parent.
+        :rtype: :py:class:`psyclone.psyir.node.Node`
+
+        '''
+        if self.parent:
+            self.parent.children.remove(self)
+        return self
+
+    def _refine_copy(self, other):
+        ''' Refine the object attributes when a shallow copy is not the most
+        appropriate operation during a call to the copy() method.
+
+        :param other: object we are copying from.
+        :type other: :py:class:`psyclone.psyir.node.Node`
+
+        '''
+        self._parent = None
+        self._has_constructor_parent = False
+        self._annotations = other.annotations[:]
+        # Invalidate shallow copied children list
+        self._children = ChildrenList(self, self._validate_child,
+                                      self._children_valid_format)
+        # And make a recursive copy of each child instead
+        self.children.extend([child.copy() for child in other.children])
+
+    def copy(self):
+        ''' Return a copy of this node. This is a bespoke implementation for
+        PSyIR nodes that will deepcopy some of its recursive data-structure
+        (e.g. the children tree), while not copying other attributes (e.g.
+        top-level parent reference).
+
+        :returns: a copy of this node and its children.
+        :rtype: :py:class:`psyclone.psyir.node.Node`
+
+        '''
+        # Start with a shallow copy of the object
+        new_instance = copy.copy(self)
+        # and then refine the elements that shouldn't be shallow copied
+        # pylint: disable=protected-access
+        new_instance._refine_copy(self)
+        return new_instance
+
+    def validate_global_constraints(self):
+        ''' Validates this Node in the context of the whole PSyIR tree.
+        Although there are validation checks for the parent<->child
+        relationships, there are other constraints that can only be
+        checked once the tree is complete and all transformations have
+        been applied. (One example is that an OMP Do directive must be
+        within the scope of an OMP Parallel directive.)
+
+        By default, this routine does nothing. It must be overridden
+        appropriately in any sub-classes to which constraints apply.
+        If an error is found then a GenerationError should be raised.
+
+        '''
+
+
+# For automatic documentation generation
+# TODO #913 the 'colored' routine shouldn't be in this module.
+__all__ = ["colored",
+           "ChildrenList",
+           "Node"]
