@@ -1,7 +1,7 @@
 # -----------------------------------------------------------------------------
 # BSD 3-Clause License
 #
-# Copyright (c) 2019-2020, Science and Technology Facilities Council.
+# Copyright (c) 2019-2021, Science and Technology Facilities Council.
 # All rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -32,7 +32,7 @@
 # POSSIBILITY OF SUCH DAMAGE.
 # -----------------------------------------------------------------------------
 # Author J. Henrichs, Bureau of Meteorology
-# Modified: A. R. Porter and S. Siso, STFC Daresbury Laboratory
+# Modified: A. R. Porter, S. Siso and R. W. Ford, STFC Daresbury Laboratory
 # -----------------------------------------------------------------------------
 
 ''' This module implementes a PSyData node, i.e.a node that at code
@@ -45,7 +45,9 @@ from psyclone.errors import InternalError
 from psyclone.f2pygen import CallGen, TypeDeclGen, UseGen
 from psyclone.psyir.nodes.statement import Statement
 from psyclone.psyir.nodes.schedule import Schedule
-from psyclone.psyir.symbols import Symbol, SymbolTable
+from psyclone.psyir.symbols import (SymbolTable, TypeSymbol, DataSymbol,
+                                    ContainerSymbol, DeferredType,
+                                    UnknownFortranType, GlobalInterface)
 
 
 # =============================================================================
@@ -93,7 +95,7 @@ class PSyDataNode(Statement):
     # Textual description of the node.
     _children_valid_format = "Schedule"
     _text_name = "PSyData"
-    _colour_key = "PSyData"
+    _colour = "green"
 
     def __init__(self, ast=None, children=None, parent=None, options=None):
 
@@ -120,10 +122,9 @@ class PSyDataNode(Statement):
             ", ".join(self.add_psydata_class_prefix(symbol) for symbol in
                       PSyDataNode.symbols)
 
-        if children:
-            # We need to store the position of the original children,
-            # i.e. before they are added to a schedule
-            node_position = children[0].position
+        # TODO #1185. Much of what is done in this constructor should either
+        # be done in a create() method or in the transformation that adds
+        # this node.
 
         # A PSyData node always contains a Schedule
         sched = self._insert_schedule(children)
@@ -138,29 +139,38 @@ class PSyDataNode(Statement):
             # FIXME: This may not be a good solution
             symtab = SymbolTable()
 
+        # Ensure that we have a container symbol for the API access
+        fortran_module = self.add_psydata_class_prefix(self.fortran_module)
+        try:
+            csym = symtab.lookup(fortran_module)
+        except KeyError:
+            csym = ContainerSymbol(fortran_module)
+            symtab.add(csym)
+
+        # The type symbol for PSyData variables
+        var_name = self.add_psydata_class_prefix("PSyDataType")
+        try:
+            type_sym = symtab.lookup(var_name)
+            if (not isinstance(type_sym.interface, GlobalInterface) or
+                    type_sym.interface.container_symbol is not csym):
+                raise InternalError(
+                    "Symbol table already contains a symbol named '{0}' but "
+                    "its interface does not refer to the '{1}' container.".
+                    format(var_name, csym.name))
+        except KeyError:
+            type_sym = TypeSymbol(var_name, DeferredType(),
+                                  interface=GlobalInterface(csym))
+            symtab.add(type_sym)
+
         # Store the name of the PSyData variable that is used for this
         # PSyDataNode. This allows the variable name to be shown in str
         # (and also, calling create_name in gen() would result in the name
         # being changed every time gen() is called).
-        self._var_name = symtab.new_symbol_name(
-            self._psy_data_symbol_with_prefix)
-        symtab.add(Symbol(self._var_name))
-
-        if children and parent:
-            # Correct the parent's list of children. Use a slice of the list
-            # of nodes so that we're looping over a local copy of the list.
-            # Otherwise things get confused when we remove children from
-            # the list.
-            for child in children[:]:
-                # Remove child from the parent's list of children
-                parent.children.remove(child)
-
-            # Add this node as a child of the parent
-            # of the nodes being enclosed and at the original location
-            # of the first of these nodes
-            parent.addchild(self, index=node_position)
-        elif parent:
-            parent.addchild(self)
+        psydata_type = UnknownFortranType(
+            "type({0}), save, target ::".format(type_sym.name))
+        self._var_name = symtab.new_symbol(
+            self._psy_data_symbol_with_prefix, symbol_type=DataSymbol,
+            datatype=psydata_type).name
 
         # Name of the region. In general at constructor time we might
         # not have a parent subroutine or any child nodes, so
@@ -175,7 +185,8 @@ class PSyDataNode(Statement):
         # query the actual name of region (e.g. during generation of a driver
         # for an extract node). If the user does not define a name, i.e.
         # module_name and region_name are empty, a unique name will be
-        # computed in gen_code(). If this name would then be stored in
+        # computed in gen_code() or lower_to_language_level(). If this name
+        # would then be stored in
         # module_name and region_name, and gen() is called again, the
         # names would not be computed again, since the code detects already
         # defined module and region names. This can then result in duplicated
@@ -321,21 +332,24 @@ class PSyDataNode(Statement):
             be added to each variable name in the post_var_list.
 
         '''
+        # Avoid circular dependency
+        # pylint: disable=import-outside-toplevel
+        from psyclone.psyGen import Kern, InvokeSchedule
         # TODO: #415 Support different classes of PSyData calls.
+        invoke = self.ancestor(InvokeSchedule).invoke
         module_name = self._module_name
         if module_name is None:
             # The user has not supplied a module (location) name so
             # return the psy-layer module name as this will be unique
             # for each PSyclone algorithm file.
-            module_name = self.root.invoke.invokes.psy.name
+            module_name = invoke.invokes.psy.name
 
         region_name = self._region_name
         if region_name is None:
             # The user has not supplied a region name (to identify
             # this particular invoke region). Use the invoke name as a
             # starting point.
-            region_name = self.root.invoke.name
-            from psyclone.psyGen import Kern
+            region_name = invoke.name
             kerns = self.walk(Kern)
             if len(kerns) == 1:
                 # This PSyData region only has one kernel within it,
@@ -424,19 +438,6 @@ class PSyDataNode(Statement):
         self._add_call("PostEnd", parent)
 
     # -------------------------------------------------------------------------
-    def gen_c_code(self, indent=0):
-        '''
-        Generates a string representation of this Node using C language
-        (currently not supported).
-
-        :param int indent: Depth of indent for the output string.
-        :raises NotImplementedError: Not yet supported for PSyData nodes.
-        '''
-        raise NotImplementedError("Generation of C code is not supported "
-                                  "for PSyDataNode.")
-
-    # -------------------------------------------------------------------------
-
     def update(self):
         # pylint: disable=too-many-branches, too-many-statements
         # pylint: disable=too-many-locals
@@ -649,4 +650,7 @@ class PSyDataNode(Statement):
         pscall = Fortran2003.Call_Stmt(reader)
         pscall.parent = fp_parent
         fp_parent.content.insert(ast_start_index, pscall)
+        # Set the pointers back into the modified parse tree
+        self.ast = pscall
+        self.ast_end = pecall
         self.set_region_identifier(routine_name, region_name)
